COLUMNS = {
    "prefix": "rhp_",
    "is_valid": "rhp_is_valid",
    "resolution": "rhp_resolution",
    "base_cell": "rhp_base_cell",
    "parent": "rhp_parent",
    "center_child": "rhp_center_child",
    "cell_area": "rhp_cell_area",
    "cell_ring": "rhp_cell_ring",
    "k_ring": "rhp_k_ring",
    "polyfill": "rhp_polyfill",
    "linetrace": "rhp_linetrace",
    "geometry": "geometry",
}
